

class Config(object):
    """User-facing Interface for the Kubernetes Configuration system.
    """
    def __init__(self, kubeconf):
        self.kubeconf = kubeconf

    def current_context(self):
        """Set the current kube context.
        """
        return self.kubeconf.get_current_context()

    def use_context(self, name):
        """Switch contexts.
        """
        self.kubeconf.set_current_context(name)


import pprint
from . import providers
from jhubctl.utils import kubectl, helm, JhubctlError


class ClusterList(object):
    """Class for managing a Kubernetes Clusters.

    This class manages your configuration for kubectl.
    """
    def __init__(self, kubeconf):
        self.kubeconf = kubeconf

    def check_cluster_exists(self, name):
        """Check if cluster exists. If it does not, raise exception."""
        self.kubeconf.open()
        clusters = self.kubeconf.get_clusters()
        names = [c['name'] for c in clusters]
        if name in names:
            return True
        return False

    def get(self, name=None, provider='AwsEKS', print_output=True):
        """List all cluster.
        """
        # Create cluster object
        Cluster = getattr(providers, provider)
        cluster = Cluster(name)

        self.kubeconf.open()
        if name is None:
            clusters = self.kubeconf.get_clusters()
            print("Running Clusters:")
            for cluster in clusters:
                print(f"  - {cluster['name']}")
        else:
            # Check that cluster exists.
            if self.check_cluster_exists(name) is False:
                raise JhubctlError("Cluster name not found in availabe clusters.")

            cluster = self.kubeconf.get_cluster(name=cluster.cluster_name)
            pprint.pprint(cluster, depth=4)

    def create(self, name, provider='AwsEKS'):
        """Create a Kubernetes cluster on a given provider.
        """
        # ----- Create K8s cluster on provider -------
        # Create cluster object
        Cluster = getattr(providers, provider)
        cluster = Cluster(name=name, ssh_key_name='zsailer')
        cluster.create()

        # -------- Add cluster to kubeconf -----------

        # Add cluster to kubeconf
        self.kubeconf.open()
        self.kubeconf.add_cluster(
            cluster.cluster_name,
            server=cluster.endpoint_url,
            certificate_authority_data=cluster.ca_cert
        )

        # Add a user to kubeconf
        self.kubeconf.add_user(name)

        # Add a user exec call for this provider.
        self.kubeconf.add_to_user(
            name,
            **cluster.kube_user_data
        )

        # Add context mapping user to cluster.
        self.kubeconf.add_context(
            name,
            cluster_name=cluster.cluster_name,
            user_name=cluster.name
        )

        # Switch contexts.
        self.kubeconf.set_current_context(name)

        # Commit changes to file.
        self.kubeconf.close()
        
        # ------ Setup autorization -------
        kubectl('apply', input=cluster.get_auth_config())

        # -------- Setup Storage ----------
        kubectl('delete', 'storageclass', 'gp2')        
        kubectl('apply', input=cluster.get_storage_config())

        # ------- setup helm locally ------
        kubectl(
            '--namespace',
            'kube-system',
            'create',
            'serviceaccount',
            'tiller'
        )

        kubectl(
            'create',
            'clusterrolebinding',
            'tiller',
            '--clusterrole=cluster-admin',
            '--serviceaccount=kube-system:tiller'
        )

        # -------- Initialize Helm -----------
        helm(
            'init',
            '--service-account',
            'tiller'
        )

        # --------- Secure Helm --------------
        kubectl(
            'patch',
            'deployment',
            'tiller-deploy',
            namespace='kube-system',
            type='json',
            patch='[{"op": "add", "path": "/spec/template/spec/containers/0/command", "value": ["/tiller", "--listen=localhost:44134"]}]'
        )

    def delete(self, name, provider='AwsEKS'):
        """Delete a Kubernetes cluster.
        """
        # if self.check_cluster_exists(name) is False:
        #     raise JhubctlError("Cluster name not found in availabe clusters.")

        # Create cluster object
        Cluster = getattr(providers, provider)
        cluster = Cluster(name)
        cluster.delete()

        # Remove from kubeconf
        self.kubeconf.open()
        self.kubeconf.remove_context(name)
        self.kubeconf.remove_user(name)
        self.kubeconf.remove_cluster(cluster.cluster_name)
        self.kubeconf.close()



from ..base import Provider

class MockProvider(Provider):
    
    provider_source = Unicode(u'mock')
    provider_alias = Unicode(u'mock')
    cluster_name = Unicode(u'')

    def check_if_cluster_is_deployed(self):
        """Returns True if the cluster is deployed and available.
        """
        return True

    def deploy(self):
        """Deploy a cluster configured for running Jupyterhub
        deployments on this provider.
        """
        pass

    def teardown(self):
        """Teardown a cluster running on this provider.
        """
        pass

    def get_authorized_users_config(self):
        """Get yaml describing authorized users for the cluster.
        """
        return None

    def get_storage_config(self):
        """Get yaml describing storage on cluster.
        """
        return None

    def get_kubectl_config(self):
        """Get yaml describing the kubeconfig for this cluster.
        """
        return None



import jinja2
import pathlib
import subprocess

class SubclassError(Exception):
    """Must be implemented in a subclass."""

class JhubctlError(Exception):
    """CLI exceptions"""


def get_flag_args(**options):
    """Build a list of flags."""
    flags = []
    for key, value in options.items():
        # Build short flags.
        if len(key) == 1:
            flag = f'-{key}'
        # Built long flags.
        else:
            flag = f'--{key}'
        flags = flags + [flag, value]
    return flags


def kubectl(*args, input=None, **flags):
    """Simple wrapper to kubectl."""
    # Build command line call.
    line = ['kubectl'] + list(args)
    line = line + get_flag_args(**flags)
    if input is not None:
        line = line + ['-f', '-']
    # Run subprocess
    output = subprocess.run(
        line,
        input=input,
        capture_output=True,
        text=True
    )
    return output


def helm(*args, input=None, **flags):
    """Simple wrapper to helm."""
    # Build command line call.
    line = ['helm'] + list(args)
    line = line + get_flag_args(**flags)
    if input is not None:
        line = line + ['-f', '-']
    # Run subprocess
    output = subprocess.run(
        line,
        input=input,
        capture_output=True,
        text=True
    )
    return output


def sanitize_path(path):
    if isinstance(path, str):
        path = pathlib.Path(path).resolve()
    elif isinstance(path, pathlib.Path):
        path = path.resolve()
    else:
        raise Exception()
    return path


def get_template(template_path, **parameters):
    """Use jinja2 to fill in template with given parameters.
    
    Parameters
    ----------
    template_path : str
        Name of template.

    Keyword Arguments
    -----------------
    Parameters passed into the jinja2 template

    Returns
    -------
    output_text : str
        Template as a string filled in with parameters.
    """
    path = sanitize_path(template_path)
    template_file = path.name
    template_dir = str(path.parent)

    ## Apply ARN of instance role of worker nodes and apply to cluster
    template_loader = jinja2.FileSystemLoader(searchpath=template_dir)
    template_env = jinja2.Environment(loader=template_loader)

    template = template_env.get_template(template_file)
    output_text = template.render(**parameters)

    return output_text



import os
import boto3
import botocore
import jinja2
import json
import subprocess
import logging
import pathlib

import tqdm

from traitlets import (
    Unicode,
    default
)
from ..base import Provider
from ....utils import get_template


CLIENT = boto3.client('cloudformation')
CREATE_WAITER = CLIENT.get_waiter('stack_create_complete')
DELETE_WAITER = CLIENT.get_waiter('stack_delete_complete')
CLOUDFORMATION = boto3.resource('cloudformation')
IAM = boto3.client('iam')
EKS = boto3.client('eks')


def stack_exists(name):
    """Use boto3 to check if a resource exists."""
    try:
        CLIENT.describe_stacks(StackName=name)
        return True
    except:
        return False


def get_stack(name):
    """Get stack from AWS's cloud formation."""
    stack_exists(name)
    stack = CLOUDFORMATION.Stack(f"{name}")
    return stack
 

def create_stack(
        stack_name,
        stack_template_path,
        parameters=None,
        capabilities=None
    ):
    # Create stack if it does not exist.
    if stack_exists(stack_name) is False:
        # Create stack
        options = {}
        if parameters is not None:
            options.update(Parameters=parameters)
        if capabilities is not None:
            options.update(Capabilities=capabilities)

        stack = CLOUDFORMATION.create_stack(
            StackName=stack_name,
            TemplateBody=get_template(stack_template_path),
            **options
        )
        # Wait for response.
        CREATE_WAITER.wait(StackName=stack.name)


def get_stack_value(stack, key):
    """Get metadata value from a cloudformation stack."""
    for output in stack.outputs:
        if output['OutputKey'] == key:
            return output['OutputValue']


def define_parameters(**parameters):
    """Get a list of parameters to pass to AWS boto call."""
    params = []
    for key, value in parameters.items():
        param = dict(ParameterKey=key, ParameterValue=value)
        params.append(param)
    return params


class AwsEKS(Provider):
    """AWS EKS configured for launching JupyterHub deployments."""
    # ------------------------------------------------------------------------
    # Configurable options
    # ------------------------------------------------------------------------

    provider_source = Unicode('Amazon Web Services EKS')
    provider_alias = Unicode('aws')

    @default('template_dir')
    def _default_template_dir(self):
        cwd = pathlib.Path(__file__).parent
        template_dir = cwd.joinpath('templates')
        return str(template_dir)

    # AWS Role NAme
    role_name = Unicode(
        help="AWS Role."
    ).tag(config=True)

    @default('role_name')
    def _default_role_name(self):
        return f'{self.name}-role'

    # Virtual private cloud name
    vpc_name = Unicode(
        help="Name of the virtual private cloud used by this deployment."
    ).tag(config=True)

    @default('vpc_name')
    def _default_vpc_name(self):
        return f'{self.name}-vpc'
        
    # Name of EKS cluster.
    cluster_name = Unicode(
        help="Name of the cluster"
    ).tag(config=True)

    @default('cluster_name')
    def _default_cluster_name(self):
        return f'{self.name}-cluster'

    # Name of node groups
    node_group_name = Unicode(
        help="Name of the node group setup to deploy jupyterhub instances."
    ).tag(config=True)

    @default('node_group_name')
    def _default_node_group_name(self):
        return f'{self.name}-node-group'

    spot_nodes_name = Unicode(
        help="Name of the spot nodes stack"
    ).tag(config=True)

    @default('spot_nodes_name')
    def _default_spot_nodes(self):
        return f'{self.name}-spot-nodes'

    utilities_name = Unicode(
        help="Name of the utilities stack"
    ).tag(config=True)

    @default('utilities_name')
    def _default_utilities_name(self):
        return f'{self.name}-utilities'

    # ------------------------------------------------------------------------
    # Provider Attributes
    # ------------------------------------------------------------------------

    @property
    def security_groups(self):
        return get_stack_value(self.vpc_stack, "SecurityGroups")

    @property
    def subnet_ids(self):
        return get_stack_value(self.vpc_stack, "SubnetIds")

    @property
    def vpc_ids(self):
        return get_stack_value(self.vpc_stack, "VpcId")

    @property
    def endpoint_url(self):
        response = EKS.describe_cluster(name=self.cluster_name)
        return response['cluster']['endpoint']

    @property
    def ca_cert(self):
        response = EKS.describe_cluster(name=self.cluster_name)
        return response['cluster']['certificateAuthority']['data']

    @property
    def node_arn(self):
        return get_stack_value(self.node_group_stack, "NodeInstanceRole")

    @property
    def node_instance_profile(self):
        return self.node_group_stack.Resource('NodeInstanceProfile').physical_resource_id

    @property
    def node_instance_role(self):
        return self.node_group_stack.Resource('NodeInstanceRole').physical_resource_id

    @property
    def node_security_group(self):
        return self.node_group_stack.Resource('NodeSecurityGroup').physical_resource_id

    @property
    def efs_id(self):
        return get_stack_value(self.utilities_stack, 'efsId')

    @property
    def admins(self):
        """Admins of the cluster."""
        return IAM.get_group(GroupName="admin")["Users"]

    # ------------------------------------------------------------------------
    # Stacks
    # ------------------------------------------------------------------------

    @property
    def role_stack(self):
        return get_stack(self.role_name)

    @property
    def vpc_stack(self):
        return get_stack(self.vpc_name)

    @property
    def node_group_stack(self):
        return get_stack(self.node_group_name)

    @property
    def spot_nodes_stack(self):
        return get_stack(self.spot_nodes_name)

    @property
    def utilities_stack(self):
        return get_stack(self.utilities_name)

    @property
    def kube_user_data(self):
        """Extra data to pass to the kubectl user for this cluster.
        
        This can be used to map extra data to clusters in the kubeconf file.
        """
        return {
            'exec': {
                'apiVersion': 'client.authentication.k8s.io/v1alpha1',
                'command': 'aws-iam-authenticator',
                'args': ["token", "-i", f"{self.cluster_name}"]
            }
        }

    # ------------------------------------------------------------------------
    # Methods
    # ------------------------------------------------------------------------

    def create(self):
        """Deploy a cluster on Amazon's EKS Service configured
        for Jupyterhub Deployments.
        """
        steps = [
            (self.create_role, (), {}),
            (self.create_vpc, (), {}),
            (self.create_cluster, (), {}),
            (self.create_node_group, (), {}),
            (self.create_spot_nodes, (), {}),
            (self.create_utilities, (), {}),
        ]
        # Execute creation.
        for step in tqdm.tqdm(steps, ncols=70):
            method, args, kwargs = step
            method(*args, **kwargs)

    def delete(self):
        """Delete a running cluster."""
        stacks = [
            self.utilities_name,
            self.spot_nodes_name,
            self.node_group_name,
            self.cluster_name,
            self.vpc_name,
            self.role_name
        ]
        # Execute creation.
        for stack in tqdm.tqdm(stacks, ncols=70):
            self.delete_stack(stack)

    def get_auth_config(self):
        """Return the Authorization Config Map (in yaml format) 
        for this cluster.
        """
        return self.get_template(
            'amazon-auth-cm.yaml',
            arn=self.node_arn,
            users=self.admins
        )

    def get_storage_config(self):
        """Return the Storage configuration (in yaml format) 
        for this cluster.
        """
        return self.get_template('amazon-storage-class.yaml')

    def get_template(self, template_name, **parameters):
        """Pull templates from the AWS templates folder"""
        template_path = pathlib.Path(self.template_dir).joinpath(template_name)
        return get_template(template_path, **parameters)

    def delete_stack(self, stack_name):
        """Teardown a stack."""
        get_stack(stack_name)
        CLIENT.delete_stack(
            StackName=stack_name
        )
        DELETE_WAITER.wait(StackName=stack_name)

    def create_stack(
        self, 
        stack_name,
        stack_template_name,
        parameters=None,
        capabilities=None
        ):
        """Create a stack using Amazon's Cloud formation"""
        # Build template_path
        stack_template_path = pathlib.Path(
            self.template_dir).joinpath(stack_template_name)

        # Create stack
        create_stack(
            stack_name,
            stack_template_path,
            parameters=parameters,
            capabilities=capabilities
        )

    def create_role(self):
        """Create an EKS Role configured to create JupyterHub Deployments
        on an EKS Provider.
        """
        self.create_stack(
            self.role_name,
            'amazon-eks-service-role.yaml',
            capabilities=['CAPABILITY_NAMED_IAM']
        )

    def create_vpc(self):
        """Create a virtual private cloud on Amazon's Web services configured
        for deploying JupyterHubs.
        """
        self.create_stack(
            self.vpc_name,
            'amazon-eks-vpc.yaml',
            parameters=define_parameters(
                VpcBlock="10.42.0.0/16",
                Subnet01Block="10.42.1.0/24",
                Subnet02Block="10.42.2.0/24",
                Subnet03Block="10.42.3.0/24"
            )
        )

    def create_cluster(self):
        """Creates a cluster on Amazon EKS .
        """
        self.create_stack(
            self.cluster_name,
            'amazon-eks-cluster.yaml',
            parameters=define_parameters(
                ClusterName=self.cluster_name,
                ControlPlaneSecurityGroup=self.security_groups,
                Subnets=self.subnet_ids
            )
        )

    def create_node_group(self):
        """Create on-demand node group on Amazon EKS.
        """
        self.create_stack(
            self.node_group_name,
            'amazon-eks-nodegroup.yaml',
            capabilities=['CAPABILITY_IAM'],
            parameters=define_parameters(
                ClusterName=self.cluster_name,
                ClusterControlPlaneSecurityGroup=self.security_groups,
                Subnets=self.subnet_ids,
                VpcId=self.vpc_ids,
                KeyName=self.ssh_key_name,
                NodeAutoScalingGroupMaxSize="1",
                NodeVolumeSize="100",
                NodeImageId="ami-0a54c984b9f908c81",
                NodeGroupName=f"{self.name} OnDemand Nodes"
            )
        )

    def create_spot_nodes(self):
        """Create spot nodes.
        """
        self.create_stack(
            self.spot_nodes_name,
            'amazon-spot-nodes.yaml',
            parameters=define_parameters(
                ClusterName=self.cluster_name,
                Subnets=self.subnet_ids,
                NodeInstanceProfile=self.node_instance_profile,
                NodeInstanceRole=self.node_instance_role,
                NodeSecurityGroup=self.node_security_group,

            )
        )

    def create_utilities(self):
        """Create utitilies stack.
        """
        self.create_stack(
            self.utilities_name,
            'amazon-utilities.yaml',
            parameters=define_parameters(
                Subnets=self.subnet_ids,
                NodeSecurityGroup=self.node_security_group
            )
        )



import secrets

from jhubctl.utils import helm, kubectl
from traitlets.config import Configurable
from traitlets import default, Unicode


class Hub(Configurable):
    """Single instance of a JupyterHub deployment.
    """
    # Helm repository to pull charts.
    helm_repo = Unicode(
        u'https://jupyterhub.github.io/helm-chart/',
        help="Jupyterhub Helm Chart repo."
    ).tag(config=True)

    # Helm chart release name. If not given, use the 
    # namespace as the release name.
    release = Unicode(
        help="Release name"
    ).tag(config=True)

    @default('release')
    def _default_release(self):
        return self.namespace

    # Helm chart version to pull.
    version = Unicode(
        u'0.7.0',
        help='Helm Chart for Jupyterhub release.'
    ).tag(config=True)

    # Name to give to namespace.
    namespace = Unicode(
        help="Name of the Kubernetes namespace."
    ).tag(config=True)

    def __init__(self, namespace, release=None, **traits):
        self.namespace = namespace
        if release is None:
            self.release = namespace
        super().__init__(**traits)

    def get_security_yaml(self):
        """Create security YAML data."""
        # Get Token.
        token = secrets.token_hex(nbytes=32)
        # Turn token into yaml
        yaml = f'proxy:\n  secretToken: "{token}"'
        return yaml

    def get(self):
        """Get specific information about this hub."""
        output = helm("get", self.release)
        if output.returncode != 0:
            print("Something went wrong!")
            print(output.stderr)
        else:
            print(output.stdout)

    def create(self):
        """Create a single instance of notebook."""
        # Point to chart repo.
        out = helm(
            "repo",
            "add",
            "jupyterhub",
            self.helm_repo
        )
        out = helm("repo", "update")

        # Get token to secure Jupyterhub
        secret_yaml = self.get_security_yaml()

        # Get Jupyterhub.
        out = helm(
            "upgrade",
            "--install",
            self.release,
            "jupyterhub/jupyterhub",
            namespace=self.namespace,
            version=self.version,
            input=secret_yaml
        )
        if out.returncode != 0:
            print(out.stderr)
        else:
            print(out.stdout)

    def delete(self):
        """Delete a Jupyterhub."""
        # Delete the Helm Release
        out = helm(
            "delete",
            self.release,
            "--purge"
        )
        if out.returncode != 0:
            print(out.stderr)
        else:
            print(out.stdout)

        # Delete the Kubernetes namespace
        out = kubectl(
            "delete",
            "namespace",
            self.namespace
        )
        if out.returncode != 0:
            print(out.stderr)
        else:
            print(out.stdout)

    def _get_description_message(self):
        """Get a description message."""
        # Describe cluster.
        out = kubectl(
            "describe",
            "services",
            "proxy-public",
            namespace=self.namespace
        )
        return out.stdout

    def _parse_description(self, description_text):
        """Turn description to dictionary."""
        text = description_text
        text = text.strip()
        lines = text.split('\n')

        data = {}
        for line in lines:
            if ":" in line:
                idx = line.index(":")
                key = line[:idx]
                value = line[idx+1:].lstrip().rstrip()
                data[key] = value
            else:
                if isinstance(value, list) is False:
                    value = [value]
                value.append(line.lstrip().rstrip())
                data[key] = value
        return data

    def get_description(self):
        """Get description (as dictionary)"""
        message = self._get_description_message()
        data = self._parse_description(message)
        return data

    def describe(self):
        """Describe jupyterhub pod."""
        print(self._get_description_message())



from ..utils import kubectl, helm
from .single import Hub

class HubList(object):
    """A class to manage many Jupyterhub deployments.

    Parameter
    ---------
    kubeconf : KubeConf object
        A KubeConf object for managing the kubeconfig
        on the current system.
    """
    def __init__(self, kubeconf):
        self.kubeconf = kubeconf

    def create(self, name):
        """Create a jupyterhub deployment on the cluster."""
        hub = Hub(namespace=name)
        hub.create()

    def get_hubs(self):
        """Get a list of hubs names.
        
        Returns
        -------
        hubs : list
            List of hub names
        """
        # Use helm to get a list of hubs.
        output = helm(
            'list',
            '-q'
        )
        # Check if an error occurred.
        if output.returncode != 0:
            print("Something went wrong!")
            print(output.stderr)
        else:
            hubs = output.stdout.split()
            return hubs

    def get(self, name=None):
        """Print a list of all jupyterHubs."""
        # Print a list of hubs.
        if name is None:
            hubs = self.get_hubs()
            print("Running Jupyterhub Deployments (by name):")
            for hub_name in hubs:
                hub = Hub(namespace=hub_name)
                data = hub.get_description()
                url = data['LoadBalancer Ingress']
                print(f'  - Name: {hub_name}')
                print(f'    Url: {url}')
        else:
            hub = Hub(namespace=name)
            hub.get()
        
    def delete(self, name):
        """Delete Hub from Kubernetes Cluster
        """
        hub = Hub(namespace=name)
        hub.delete()

    def describe(self, name):
        """Describe a cluster."""
        hub = Hub(namespace=name)
        hub.describe()



import pathlib
from ...utils import SubclassError
from traitlets.config import Configurable
from traitlets import (
    Unicode,
    Dict,
    default
)

class Provider(Configurable):
    """Base class for Kubernetes Cluster providers.

    To create a new provider, inherit this class and 
    replace the folowing traits and methods with the 
    logic that is appropriate for the provider.

    We recommend creating a new folder for that provider
    where all templates can be grouped.
    """
    # Specific type of the provider
    provider_type = Unicode(help="Provider type")

    # An alias for the provider. No spaces or hyphens. Underscores instead.
    provider_alias = Unicode(help="Simple alias pointing to this provider.")

    # 
    cluster_name = Unicode(help="Name of cluster.")

    # Path to templates for this provider.
    template_dir = Unicode(
        help="Path to template"
    ).tag(config=True)

    ssh_key_name = Unicode(
        help='User SSH key name'
    ).tag(config=True)

    @property
    def kube_user_data(self):
        """Extra data to pass to the kubectl user for this cluster.
        
        This can be used to map extra data to clusters in the kubeconf file.
        """
        return None

    def __init__(self, name, **traits):
        self.name = name
        super().__init__(**traits)

    def check_if_cluster_is_deployed(self):
        """Returns True if the cluster is deployed and available.
        """
        raise SubclassError("Must be implemented in a subclass.")

    def create(self):
        """Deploy a cluster configured for running Jupyterhub
        deployments on this provider.
        """
        raise SubclassError("Must be implemented in a subclass.")

    def delete(self):
        """Teardown a cluster running on this provider.
        """
        raise SubclassError("Must be implemented in a subclass.")

    def get_auth_config(self):
        """Get yaml describing authorized users for the cluster.
        """
        raise SubclassError("Must be implemented in a subclass.")

    def get_storage_config(self):
        """Get yaml describing storage on cluster.
        """
        raise SubclassError("Must be implemented in a subclass.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import sys
from shutil import rmtree

from setuptools import setup, find_packages

# Package meta-data.
NAME = 'jhubctl'
DESCRIPTION = 'CLI for managing Jupyterhub instances on AWS+Kubernetes'
URL = 'https://github.com/townsenddw/jhubctl'
EMAIL = ''
AUTHOR = 'Dwight Townsend, Zach Sailer'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = None

REQUIRED = [
    "pyyaml",
    "traitlets_paths",
    "traitlets"
    "tqdm",
    "ipython",
    "kubeconf",
    "boto3",
    "jinja2",
]

setup(
    name=NAME,
    version="0.0.1",
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    entry_points={
        'console_scripts': ['jhubctl=jhubctl.main:main'],
    },
    install_requires=REQUIRED,
    include_package_data=True,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
)



import sys
import os
import re
from copy import deepcopy 

from ipython_genutils.text import indent
from ipython_genutils import py3compat

from traitlets.config.loader import (
    KVArgParseConfigLoader, PyFileConfigLoader, Config, ArgumentError, ConfigFileNotFound, JSONFileConfigLoader
)
from traitlets.config import (
    Application, 
    Configurable, 
    catch_config_error
)

from traitlets import (
    default,
    observe,
    Unicode,
    List,
    Dict
)

from kubeconf import KubeConf

from .utils import JhubctlError
from .clusters import providers, ClusterList
from .hubs import HubList, Hub


def exception_handler(exception_type, exception, traceback):
    """Handle jhubctl exceptions"""
    print(f'{exception_type.__name__}: {exception}')

#sys.excepthook = exception_handler


class JhubctlApp(Application):
    """A traitlets application that deploys jupyterhub on Kubernetes clusters.

    Example call:
        $ jhubctl create hub "my_hub" 

    JhubctlApp manages both clusters and jupyterhub deployments through
    a single interface. 

    Subcommands: 
    
        $ jhubctl get <resource> <name> : List a named resource found in kubeconfig.
        $ jhubctl get <resource> : List all resources found in kubeconfig.
        $ jhubctl create <resource> <name> : Create a resource with the given name.
        $ jhubctl delete <resource> <name> : Delete a resource with the given name.
    

    JhubctlApp is configurable through traitlets config system. Configurable traits
    can be set at the command line or in a config file. To generate a config 
    template, use the following flag:

        $ jhubctl --generate-config
    """
    # Name of the commandline application
    name = Unicode(u'jhubctl')

    # Documentation used at the top of the CLI help.
    description = Unicode(
        u"A familiar CLI for managing JupyterHub "
        u"deployments on Kubernetes clusters."
    )

    # Printed examples for the help.
    examples = Unicode(
        u" > jhubctl create hub myhub"
    )

    # Classes to expose to the config system
    classes = List([
        KubeConf,
        Hub
    ])

    # Provider to configure.
    provider_type = Unicode(
        u'AwsEKS',
        help="Provider type."
    ).tag(config=True)
    
    # Subcommands allowed by application.
    subcommands = Dict({
        'create': ((), 'Create a resource.'),
        'delete': ((), 'Delete a resource.'),
        'get': ((), 'List a resource or resources'),
        'describe': ((), 'Describe a resource')
    })

    # Resource that can be deployed and managed.
    resources = List([
        'cluster',
        'hub',
    ])

    # Name of the configuration file to read.
    config_file = Unicode(
        help="Name of configuration file."
    ).tag(config=True)

    @default('config_file')
    def _default_config_file(self):
        return u"jhubctl_config.py"

    def print_subcommands(self):
        """Print the subcommand part of the help."""
        lines = ["Call"]
        lines.append('-'*len(lines[-1]))
        lines.append('')
        lines.append("> jhubctl <subcommand> <resource-type> <resource-name>")
        lines.append('')
        lines.append("Subcommands")
        lines.append('-'*len(lines[-1]))
        lines.append('')

        for name, subcommand in self.subcommands.items():
            lines.append(name)
            lines.append(indent(subcommand[1]))

        lines.append('')
        print(os.linesep.join(lines))

    @catch_config_error
    def parse_command_line(self, argv=None):
        """Parse the jhubctl command line arguments.
        
        This overwrites traitlets' default `parse_command_line` method
        and tailors it to jhubctl's needs.
        """
        argv = sys.argv[1:] if argv is None else argv
        self.argv = [py3compat.cast_unicode(arg) for arg in argv]

        # Append Provider Class to the list of configurable items.
        ProviderClass = getattr(providers, self.provider_type)
        self.classes.append(ProviderClass)

        if any(x in self.argv for x in ('-h', '--help-all', '--help')):
            self.print_help('--help-all' in self.argv)
            self.exit(0)

        if '--version' in self.argv or '-V' in self.argv:
            self.print_version()
            self.exit(0)

        # Generate a configuration file if flag is given.
        if '--generate-config' in self.argv:
            conf = self.generate_config_file()
            with open(self.config_file, 'w') as f:
                f.write(conf)
            self.exit(0)

        # If not config, parse commands.
        ## Run sanity checks.
        # Check that the minimum number of arguments have been called.
        if len(self.argv) < 2:
            raise JhubctlError(
                "Not enough arguments. \n\n"
                "Expected: jhubctl <action> <resource> <name>")

        # Check action
        self.resource_action = self.argv[0]
        if self.resource_action not in self.subcommands:
            raise JhubctlError(
                f"Subcommand is not recognized; must be one of these: {self.subcommands}")

        # Check resource
        self.resource_type = self.argv[1]
        if self.resource_type not in self.resources:
            raise JhubctlError(
                f"First argument after a subcommand must one of these"
                f"resources: {self.resources}"
            )

        # Get name of resource.
        try:
            self.resource_name = self.argv[2]
        except IndexError:
            if self.resource_action != "get":
                raise JhubctlError(
                    "Not enough arguments. \n\n"
                    "Expected: jhubctl <action> <resource> <name>")
            else:
                self.resource_name = None

        # flatten flags&aliases, so cl-args get appropriate priority:
        flags, aliases = self.flatten_flags()
        loader = KVArgParseConfigLoader(argv=argv, aliases=aliases,
                                        flags=flags, log=self.log)
        config = loader.load_config()
        self.update_config(config)
        # store unparsed args in extra_args
        self.extra_args = loader.extra_args

    def initialize(self, argv=None):
        """Handle specific configurations."""
        # Parse configuration items on command line.
        self.parse_command_line(argv)
        if self.config_file:
            self.load_config_file(self.config_file)

        # Initialize objects to interact with.
        self.kubeconf = KubeConf()
        self.cluster_list = ClusterList(kubeconf=self.kubeconf)
        self.hub_list = HubList(kubeconf=self.kubeconf)

    def start(self):
        """Execution happening on jhubctl."""
        # Get specified resource.
        resource_list = getattr(self, f'{self.resource_type}_list')
        resource_action = getattr(resource_list, self.resource_action)
        resource_action(self.resource_name)


def main():
    app = JhubctlApp()
    app.initialize()
    app.start()


if __name__ == "__main__":
    main()

